;;; <a-stern.scm>
; a* (nicht iterativ) beim kürzesten Weg
(require (lib "breakpoint.ss"))
; Datenstruktur Feld: (x-koordinate . y-koordinate)
;;; ===== Kontrollstrukturen für die Orte und das Feld
; Start und Ziel
(define start '(7 . 4))
(define ziel '(1 . 3))
; der Rand
(define
  (im-feld? ort)
  (and
   (> (car ort) 0)
   (< (car ort) 11)
   (> (cdr ort) 0)
   (< (cdr ort) 11)))
; die Hindernisse
(define
  hindernisse-assoziationsliste
  '((1 2 6)                      ; (zeilenindex spaltenindex-1 ...)
    (2 2 3 4 6)
    (3 2 6 8 9)
    (4 6 8 9)
    (5 9)
    (6 3 4 5 6)
    (7 3 7)
    (8 3 7)
    (9 3 4)
    (10 6 7)))
; Prädikat dazu
(define
  (ist-hindernis? ort)
  (member
   (cdr ort)
   (cdr (assoc (car ort) hindernisse-assoziationsliste))))
; Filterfunktion für zulässige Koordinaten
(define
  (zulaessig? ort)
  (and
   (im-feld? ort)
   (not (ist-hindernis? ort))))
;;; ===== nachfolger =====
; erzeugt alle zulässigen Nachfolger eines Feldes
(define
  (nachfolger ort)
  (append
   (if (zulaessig? (cons (sub1 (car ort)) (cdr ort))) (list (cons (sub1 (car ort)) (cdr ort))) '())
   (if (zulaessig? (cons (add1 (car ort)) (cdr ort))) (list (cons (add1 (car ort)) (cdr ort))) '())
   (if (zulaessig? (cons (car ort) (sub1 (cdr ort)))) (list (cons (car ort) (sub1 (cdr ort)))) '())
   (if (zulaessig? (cons (car ort) (add1 (cdr ort)))) (list (cons (car ort) (add1 (cdr ort)))) '())))
;;; ===== Restkostenfunktion =====
; Berechnet die Manhattan - Distanz der Felder ort und ziel.
(define
  (restkosten ort ziel)
  (+
   (abs
    (- (car ort) (car ziel)))
   (abs
    (- (cdr ort) (cdr ziel)))))
;;; ===== Kostenfunktion =====
(define
  (kosten weg ziel)
  (+ (length weg) (restkosten (car weg) ziel)))

;;; ===== expandiere Knoten =====
; es wird das erste Feld des ersten Weges mit minimalen Kosten expandiert.
; ----- 1: keine nachfolger (mehr) -----
; wenn das ziel unter den Feldern ist, werden die anderen verworfen.
(define
  (expandiere weg ziel)
  (let lokal ((nachfolger (nachfolger (car weg))) (akku'()))
    (cond
      [(null? nachfolger) akku]
      [else
       'wert-undefiniert])))
;Test:
(writeln "expandiere 1:")
(expandiere (list start) ziel)
(expandiere (list (cons -10 -10)) ziel)

; ----- 2: ziel enthalten -----
; wenn das ziel unter den Feldern ist, werden die anderen verworfen.
(define
  (expandiere weg ziel)
  (let lokal ((nachfolger (nachfolger (car weg))) (akku'()))
    (cond
      [(null? nachfolger) akku]
      [(member ziel nachfolger)
       (list (cons (car (member ziel nachfolger)) weg))]
      [else
       'wert-undefiniert])))
;Test:
(writeln "expandiere 2:")
(expandiere (list start) ziel)
(expandiere (list (cons 1  5)) ziel)

; ----- 3: da war ich schon -----
(define
  (expandiere weg ziel)
  (let lokal ((nachfolger (nachfolger (car weg))) (akku'()))
    (cond
      [(null? nachfolger) akku]
      [(member ziel nachfolger)
       (list (cons (car (member ziel nachfolger)) weg))]
      [(member (car nachfolger) weg)              
       (lokal (cdr nachfolger) akku)]
      [else
       (lokal (cdr nachfolger) (cons (cons (car nachfolger) weg) akku))])))
;Test:
(writeln "expandiere 3:")
(expandiere (list start) ziel)
(expandiere (list (cons 1 5)) ziel)
(expandiere (list (cons 2 5) (cons 3 5)) ziel)

; ----- 4: sonst weiter -----
(define
  (expandiere weg ziel)
  (let lokal ((nachfolger (nachfolger (car weg))) (akku'()))
    (cond
      [(null? nachfolger) akku]
      [(member ziel nachfolger)
       (list (cons (car (member ziel nachfolger)) weg))]
      [(member (car nachfolger) weg)              
       (lokal (cdr nachfolger) akku)]
      [else
       (lokal (cdr nachfolger) (cons (cons (car nachfolger) weg) akku))])))
;Test:
(writeln "expandiere 4:")
(expandiere (list start) ziel)
(expandiere (list (cons 1  5)) ziel)

;;; ===== fuege-ein: Verwaltung der Prioritätswarteschlange =====
; die expandierten neuen Wege werden nach den Schätzkosten
; in eine Prioritätswarteschlange der bisherigen Wege eingefügt.
; der alte Weg (am Kopf) wird extern entfernt.
; ----- alle eingefügt -----
(define
  (fuege-ein neue-wege wege wege-merken ziel)
  (cond
    [(null? neue-wege)
     wege]
    [else
     'wert-undefiniert]))

;Test:
(writeln "fuege-ein 1:")
(fuege-ein '() (list start) '() ziel)
(fuege-ein (expandiere (list start) ziel) '() '() ziel)
; ----- keine wege (mehr) -----
; neu starten mit den nächsten neuen Weg.
(define
  (fuege-ein neue-wege wege wege-merken ziel)
  (cond
    [(null? neue-wege)
     wege]
    [(null? wege)
     (fuege-ein (cdr neue-wege) (reverse (cons (car neue-wege) wege-merken)) '() ziel)]
    [else
     'wert-undefiniert]))

;Test:
(writeln "fuege-ein 2:")
(fuege-ein (expandiere (list start) ziel) '() '() ziel)
; ----- passt dort hinein -----
(define
  (fuege-ein neue-wege wege wege-merken ziel)
  (cond
    [(null? neue-wege)
     wege]
    [(null? wege)
     (fuege-ein (cdr neue-wege) (reverse (cons (car neue-wege) wege-merken)) '() ziel)]
    [(>= (kosten (car wege) ziel) (kosten (car neue-wege) ziel))
     (fuege-ein (cdr neue-wege) (append (reverse wege-merken)(cons (car neue-wege) wege)) '() ziel)]
    [else
     'wert-undefiniert]))

;Test:
(writeln "fuege-ein 3:")
(fuege-ein (expandiere (list start) ziel) '() '() ziel)
; ----- Position noch nicht gefunden -----
; sonst - Fall
(define
  (fuege-ein neue-wege wege wege-merken ziel)
  (cond
    [(null? neue-wege)
     wege]
    [(null? wege)
     (fuege-ein (cdr neue-wege) (reverse (cons (car neue-wege) wege-merken)) '() ziel)]
    [(>= (kosten (car wege) ziel) (kosten (car neue-wege) ziel))
     (fuege-ein (cdr neue-wege) (append (reverse wege-merken)(cons (car neue-wege) wege)) '() ziel)]
    [else
     (fuege-ein neue-wege (cdr wege) (cons (car wege) wege-merken) ziel)]))

;Test:
(writeln "fuege-ein 4:")
(fuege-ein (expandiere (list start) ziel) '() '() ziel)

; ----- entferne-mehrfachwege -----
; alle dem weg nachfolgenden Wege mit dem selben Kopfelement können aus der
; wege-liste entfernt werden.
(define
  (entferne-mehrfachwege wege)
  (if
   (null? wege)
   wege
   (cons
    (car wege)
    (let loop
      ((erster-kopf (caar wege))
       (restliche-wege (cdr wege)))
      (cond
        ((null? restliche-wege) restliche-wege)
        ((equal? erster-kopf (caar restliche-wege))
         (loop erster-kopf (cdr restliche-wege)))
        (else (cons (car restliche-wege)(loop erster-kopf (cdr restliche-wege)))))))))

;;; ===== a* - Verfahren selbst =====
; Der entscheidende Punkt ist, dass immer nur Wege mit den minimalen Schätzkosten
; weiter entwickelt werden, so dass beim Auffinden des ersten vollständigen Weges
; genau dann ein Weg mit minimalen Kosten gefunden wurde, wenn die Schätzfunktion
; für die Restkosten optimistisch war. 
; Das funktioniert mit einer Prioritätswarteschlange.
; Zusätzlich könnten die nachfolgenden Wege um die bereinigt werden, die dasselbe 
; Element wie das Kopfelement enthalten. --> fehlt noch!!! (s.o.)
; ----- Schritt 1: Kopf der Funktion -----
(define
  (a* start ziel wege)
  'wert-undefiniert)
;Test:
(writeln "a* 1:")
(a* start ziel (list (list start)))

; ----- Schritt 2: Keine Wege (mehr) -----
(define
  (a* start ziel wege)
  (cond
    [(null? wege)
     'nicht-loesbar]
    [else
     'wert-undefiniert]))
;Test:
(writeln "a* 2:")
(a* start ziel '())
(a* start ziel (list (list start)))

; ----- Schritt 3: am Ziel -----
; da das Expandieren so arbeitet, dass beim Erreichen des Ziels dieses am Kopf
; des Kopfelementes der Wegeliste steht, kann darauf getestet werden.
(define
  (a* start ziel wege)
  (cond
    [(null? wege)
     'nicht-loesbar]
    [(member ziel (car wege))
     (car wege)]
    [else
     'wert-undefiniert]))
;Test:
(writeln "a* 3:")
(a* start ziel '())
(a* start ziel (list (list ziel start)))
(a* start ziel (list (list start)))

; ----- Schritt 4: sonst weiter -----
(define
  (a* start ziel wege)
  (cond
    [(null? wege)
     'nicht-loesbar]
    [(member ziel (car wege))
     (reverse (car wege))]
    [else
     (a*
      start
      ziel
      (entferne-mehrfachwege 
       (fuege-ein (expandiere (car wege) ziel) (cdr wege) '() ziel)))]))
; ----- zeige-feld -----
; Hilfsfunktion zur Darstellung des Feldes:
(define
  (zeige-feld x y)
  (display " ")
  (cond
    ((> x 10) (display #\newline))
    ((> y 10)
     (display #\newline)(zeige-feld (+ x 1) 1))
    ((ist-hindernis? (cons x y))
     (display "O")(zeige-feld x (+ y 1)))
    (else
     (display "+")(zeige-feld x (+ y 1)))))

;Test:
(writeln "a* 4:")
(a* start ziel (list (list start)))

;;; ===== interessant auch: =====
(a* start (cons 2 9) (list (list start)))
(a* start (cons 2 8) (list (list start)))

(zeige-feld 1 1)
