;;; a-stern-2010.scm
(require (lib "breakpoint.ss"))
; 
; Generieren des Feldes ...
; ist nicht notwendig. Es reicht ein ...
;;; ===== unzulaessig? =====
; Praedikat zum Pruefen, ob eine Position zulaessig ist.
; Datenstruktur Feld: ((zeilenindex spaltenindex) ...
; Indices jeweils von 1 bis 10.
; Zunaechst werden alle unzulaessigen Positionen definiert.
; ----- die Hindernisse -----
(define
  hindernisse-assoziationsliste
  '((1 2 6)                      ; (zeilenindex spaltenindex-1 ...)
    (2 2 3 4 6)
    (3 2 6 8 9)
    (4 6 8 9)
    (5 9)
    (6 3 4 5 6)
    (7 3 7)
    (8 3 7)
    (9 3 4)
    (10 6 7)))
; ----- und der Rand -----
(define
  (unzulaessig? position)
  (or
   (< (first position) 1)
   (> (first position) 10)
   (< (second position) 1)
   (> (second position) 10)
   (member
    (second position)
    (rest
     (assoc
      (first position)
      hindernisse-assoziationsliste)))))

;;; ===== alle-nachfolger =====
; liefert eine Liste aller zulaessigen Nachfolgepositionen zurueck
; benoetigt unzulaessig?
(define
  (alle-nachfolger position)
  (let loop
    ((moeglichkeiten '((1 0) (-1 0) (0 1) (0 -1))))
    (cond
      ((null? moeglichkeiten) '())
      ((unzulaessig? (map + (first moeglichkeiten) position))
       (loop (rest moeglichkeiten)))
      (else
       (cons
        (map + (first moeglichkeiten) position)
        (loop (rest moeglichkeiten)))))))
; Tests:(alle-nachfolger '(1 3))(alle-nachfolger '(4 4))


;;; ===== alle-neuen-wege =====
; baut damit die neuen Wege
; benoetigt alle-nachfolger
(define
  (alle-neuen-wege weg)
  (let loop
    ((positionen (alle-nachfolger (first weg))))
    (cond
      ((null? positionen) '())
      (else
       (cons
        (cons (first positionen) weg)
        (loop (rest positionen)))))))

; Es werden die drei Funktionen zur Bestimmung der Kosten benoetigt
;;; ===== aktuelle-kosten =====
; berechnet die aktuellen Kosten des Weges
(define
  (aktuelle-kosten weg)
  (- (length weg) 1))
;;; ===== restkosten =====
; berechnet die geschaetzten Restkosten der aktuellen Position;
; fuer die Manhattan-Distanz benoetigt die Funktion das ziel.
(define
  (restkosten position ziel)
  (apply + (map abs (map - position ziel))))
; Test:(restkosten '(4 9) '(1 3))(restkosten '(9 5) '(1 7))

;;; ===== gesamtkosten =====
; berechnet die geschaetzten Gesamtkosten des Weges,
; das ist die Summe aus den aktuellen Kosten und den geschaetzten Restkosten
(define
  (gesamtkosten weg ziel)
  (+ (aktuelle-kosten weg) (restkosten (first weg) ziel)))

; ----- PrioWS -----
; Auch hier ist eine Prioritätswarteschlange nach den Bewertungen -hier den geschaetzten Gesamtkosten- zu verwalten.
;Die Elemente sehen etwas anders aus, das Prinzip bleibt. Daher:
; benötigt ein Praedikat kuerzer?
(define
  (kuerzer? weg-1 weg-2 ziel)
  (< (gesamtkosten weg-1 ziel) (gesamtkosten weg-2 ziel)))

(define
  (fuege-einen-weg-ein weg priows ziel)
  (cond
    ((null? priows) (list weg))
    ((kuerzer? weg (first priows) ziel)
     (cons weg priows))
    (else
     (cons
      (first priows)
      (fuege-einen-weg-ein weg (rest priows) ziel)))))
; wird verwendet von:
(define
  (fuege-alle-wege-ein wege priows ziel)
  (cond
    ((null? wege) priows)
    (else
     (fuege-alle-wege-ein
      (rest wege)
      (fuege-einen-weg-ein (first wege) priows ziel)
      ziel))))

;;; ===== entferne-teurere =====
; aus den neu aufzunehmenden Wegen muessen die Wege entfernt werden, die am Kopf dieselbe Position haben
; wie ein schon vorhandener aus der PrioWS, aber teurer sind und ...
; aus der PrioWS muessen die Wege entfernt werden, die am Kopf dieselbe Position haben, aber teurer sind;
; erst danach darf eingefuegt werden.
(define
  (entferne-teurere-neue wege priows ziel)
  (cond
    ((null? wege) priows)
    ((not (assoc (first (first wege)) priows))
     ; der muss rein
     (fuege-einen-weg-ein
      (first wege)
      (entferne-teurere-neue (rest wege) priows ziel)
       ziel))
    ((kuerzer? (first wege) (assoc (first (first wege)) priows) ziel)
      ;der muss rein und der alte muss raus!
     (fuege-einen-weg-ein
      (first wege)
      (entferne-teurere-neue (rest wege) (entferne-teureren-alten (first wege) priows) ziel)
      ziel))
    (else
     ;der darf nicht rein
     (entferne-teurere-neue (rest wege) priows ziel))))
; REDESIGN!
(define
  (entferne-teureren-alten weg priows)
  (append
   (reverse (rest (assoc (first weg) (reverse priows))))
   (rest (assoc (first weg) priows)) ; hintere
   ))

;;; ===== a-stern =====
; die Funktion a-stern selbst fuehrt die einzelnen Schritte solange durch,
; bis die aktuelle Position das Ziel ist.
(define
  (a-stern priows ziel)
  (cond
    ((equal? (first (first priows)) ziel) (first priows))
    (else
     (a-stern (entferne-teurere-neue (alle-neuen-wege (first priows)) (rest priows) ziel) ziel))))
; Sie wird mit dem Weg aufgerufen, der allein die Startposition enthaelt.
(a-stern '(((9 6))) '(1 4))